#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

install_requires = \
['opentelemetry-instrumentation', 'opentelemetry-sdk']

extras_require = \
{'test': ['pytest', 'pytest-cov', 'pytest-html']}

setup(name='otelme',
      version='0.0.4',
      description='otelme pronounced "Oh, tell me"',
      author=None,
      author_email='Ryan Scott Brown <otelme@ryansb.com>',
      url=None,
      py_modules=['otelme'],
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.8',
     )
